/*******************************************************************************
 * Copyright 2015 - 2019 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.bsh.web;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpHeaders;
import org.htmlunit.FailingHttpStatusCodeException;
import org.htmlunit.HttpMethod;
import org.htmlunit.NicelyResynchronizingAjaxController;
import org.htmlunit.Page;
import org.htmlunit.UnexpectedPage;
import org.htmlunit.WebClient;
import org.htmlunit.WebClientOptions;
import org.htmlunit.WebRequest;
import org.htmlunit.util.NameValuePair;

import com.fasterxml.jackson.databind.ObjectMapper;

import de.frankmuenster.mahoe.banks.bsh.SchwaebischHall;
import de.frankmuenster.mahoe.banks.bsh.json.Account;
import de.frankmuenster.mahoe.banks.bsh.json.AccountDetailsResponse;
import de.frankmuenster.mahoe.banks.bsh.json.AccountListResponse;
import de.frankmuenster.mahoe.banks.bsh.json.Error;
import de.frankmuenster.mahoe.banks.bsh.json.Exemption;
import de.frankmuenster.mahoe.banks.bsh.json.ExemptionsResponse;
import de.frankmuenster.mahoe.banks.bsh.json.IbanResponse;
import de.frankmuenster.mahoe.banks.bsh.json.Message;
import de.frankmuenster.mahoe.banks.bsh.json.MessageCountResponse;
import de.frankmuenster.mahoe.banks.bsh.json.MessageReadResponse;
import de.frankmuenster.mahoe.banks.bsh.json.MessagesResponse;
import de.frankmuenster.mahoe.banks.bsh.json.Status;
import de.frankmuenster.mahoe.banks.bsh.json.accountdetails.AccountDetailData;
import de.frankmuenster.mahoe.banks.bsh.json.accountdetails.AccountDetailMapper;
import de.frankmuenster.mahoe.messages.MahoeMessage;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.utils.WebUtilsHtmlUnit;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;

/**
 * Utility-Klasse fr die BSH Website
 *
 * @author Frank Mnster
 */
@SuppressWarnings({ "PMD.ExcessiveImports", "PMD.GodClass", "PMD.TooManyMethods" })
public final class BshWebSiteUtils {

    /**
     * Error message for an unknown/empty error in JSON response
     */
    private static final String UNKOWN_ERROR = I18N.tr("_unknownError"); //$NON-NLS-1$

    /**
     * Jackson Mapper
     */
    private static final ObjectMapper MAPPER = new ObjectMapper();

    /**
     * MessageRead ID POST-Parameter
     */
    private static final String PARAM_ID = "id"; //$NON-NLS-1$

    /**
     * MessageRead Command POST-Parameter
     */
    private static final String PARAM_COMMAND = "command"; //$NON-NLS-1$

    /**
     * MessageRead Command
     */
    private static final String COMMAND_READ = "readmessage"; //$NON-NLS-1$

    /**
     * Liefert die allg. Vertragsdaten zu den gegebenen VTN.
     *
     * @param webClient
     * @param vtn
     * @return die allg. Vertragsdaten.
     * @throws WebSiteException
     *             , falls kein Vertrag mit der angegebenen VTN gefunden wurde.
     */
    public static Account getAccount(final WebClient webClient, final String vtn) throws WebSiteException {

        final List<Account> accounts = BshWebSiteUtils.getAccounts(webClient);
        Account account = null;
        for (final Account acc : accounts) {
            if (acc.getVtn().equals(vtn)) {
                account = acc;
                break;
            }
        }
        if (account == null) {
            throw new WebSiteException(I18N.tr("BshWebSiteUtils.contractNotExists", vtn), //$NON-NLS-1$
                    WebSiteExceptionTypes.ACCOUNT_NOT_FOUND);
        }
        return account;
    }

    /**
     * Liefert das AccountDetailObject mit allen Angaben zu dem Vertrag.
     *
     * @param webClient
     * @param vtn
     *            die VTN Nummer des Vertrages.
     * @return das AccountDetailObject
     * @throws WebSiteException
     */
    public static AccountDetailData getAccountDetails(final WebClient webClient, final String vtn)
            throws WebSiteException {

        AccountDetailData accountDetails = null;
        try {
            final String url = SchwaebischHall.URL_DETAILS + "?vtn=" + vtn; //$NON-NLS-1$
            final Page responsePage = webClient.getPage(url);
            if (WebUtilsHtmlUnit.webResponseIsJson(responsePage.getWebResponse())) {

                final UnexpectedPage pape = (UnexpectedPage) responsePage;
                Logger.debug("GetAccountDetails RWebResponseCode :" + pape.getWebResponse().getStatusCode()); //$NON-NLS-1$
                Logger.debug(responsePage.getWebResponse().getContentAsString());

                final AccountDetailsResponse deatilsResponse = BshWebSiteUtils.MAPPER
                        .readValue(responsePage.getWebResponse().getContentAsStream(), AccountDetailsResponse.class);
                if (deatilsResponse.getStatus() == Status.ERROR) {
                    final List<Error> errors = deatilsResponse.getErrors();
                    for (final Error error : errors) {
                        Logger.error(error.getMessage());
                    }
                    String errorText = BshWebSiteUtils.UNKOWN_ERROR;
                    if (!errors.isEmpty()) {
                        errorText = errors.get(0).getMessage();
                    }
                    throw new WebSiteException(errorText, WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
                }

                accountDetails = deatilsResponse.getAccountDetails();

            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(e.getMessage());
            throw new WebSiteException(I18N.tr("BshWebSiteUtils.accountDetailsRequestFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }

        return accountDetails;
    }

    /**
     * Get a list of all account for the logged in user.
     *
     * @param webClient
     *            with already authenticated user
     * @return the list of all accounts for the user
     * @throws WebSiteException
     */
    public static List<Account> getAccounts(final WebClient webClient) throws WebSiteException { // NOPMD

        List<Account> accounts = new ArrayList<>();
        try {
            final Page responsePage = webClient.getPage(SchwaebischHall.URL_CONTRACTS);
            if (WebUtilsHtmlUnit.webResponseIsJson(responsePage.getWebResponse())) {

                final UnexpectedPage page = (UnexpectedPage) responsePage;
                Logger.debug("GetAccounts WebResponseCode :" + page.getWebResponse().getStatusCode()); //$NON-NLS-1$
                Logger.debug(responsePage.getWebResponse().getContentAsString());

                final AccountListResponse accountResponse = BshWebSiteUtils.MAPPER
                        .readValue(responsePage.getWebResponse().getContentAsStream(), AccountListResponse.class);
                if (accountResponse.getStatus() == Status.ERROR) {
                    final List<Error> errors = accountResponse.getErrors();
                    for (final Error error : errors) {
                        Logger.error(error.getMessage());
                    }
                    String errorText = BshWebSiteUtils.UNKOWN_ERROR;
                    if (!errors.isEmpty()) {
                        errorText = errors.get(0).getMessage();
                    }
                    throw new WebSiteException(errorText, WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
                }
                accounts = accountResponse.getAccounts();
            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(e.getMessage());
            throw new WebSiteException(I18N.tr("BshWebSiteUtils.accountRequestFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }

        return accounts;
    }

    /**
     * Liefert eine Liste mit Ums&auml;tzen aus den Vertragsdetails-Objekt.
     *
     * @param accountDetailData
     *            das Vertragsdetails-Objekt
     * @return Liste mit Hibiscus-Ums&auml;tzen
     */
    public static List<Umsatz> getBookedTransctions(final AccountDetailData accountDetailData) {

        List<Umsatz> umsaetze = null;
        final AccountDetailMapper mapper = new AccountDetailMapper(accountDetailData);
        umsaetze = mapper.getUmsatzList();

        return umsaetze;
    }

    /**
     * Liefert den Freistellungsauftrag zur CoreID.
     *
     * @param webClient
     * @param coreId
     * @return den Freistellungsauftrag, oder <i>NULL</i>
     * @throws WebSiteException
     */
    public static Exemption getExemption(final WebClient webClient, final String coreId) throws WebSiteException {

        Exemption result = null;
        final List<Exemption> exemptions = BshWebSiteUtils.getExemptions(webClient);
        for (final Exemption exemption : exemptions) {
            if (coreId.equals(exemption.getCoreid())) {
                result = exemption;
                break;
            }
        }
        return result;
    }

    /**
     * Liefert eine Liste mit den Freistellungsantr&auml;gen.
     *
     * @param webClient
     *            das angemeldete WebClient Objekt.
     * @return eine Liste mit den Freistellungsantr&auml;gen.
     * @throws WebSiteException
     */
    public static List<Exemption> getExemptions(final WebClient webClient) throws WebSiteException { // NOPMD

        List<Exemption> exemptions = new ArrayList<>();
        try {
            final Page responsePage = webClient.getPage(SchwaebischHall.URL_EXEMPTION);
            if (WebUtilsHtmlUnit.webResponseIsJson(responsePage.getWebResponse())) {

                final UnexpectedPage page = (UnexpectedPage) responsePage;
                Logger.debug("GetAccounts WebResponseCode :" + page.getWebResponse().getStatusCode()); //$NON-NLS-1$
                Logger.debug(responsePage.getWebResponse().getContentAsString());

                final ExemptionsResponse exemptionResponse = BshWebSiteUtils.MAPPER
                        .readValue(responsePage.getWebResponse().getContentAsStream(), ExemptionsResponse.class);
                if (exemptionResponse.getStatus() == Status.ERROR) {
                    BshWebSiteUtils.handleExemptionError(exemptionResponse);
                } else {
                    exemptions = exemptionResponse.getExemptions();
                }
            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(e.getMessage());
            throw new WebSiteException(I18N.tr("BshWebSiteUtils.accountRequestFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }

        return exemptions;
    }

    /**
     * Liefert die Anzahl neuer ungelesener Nachrichten.
     *
     * @param webClient
     *            der angemeldete WebClient
     * @return die Anzahl neuer ungelesener Nachrichten
     * @throws WebSiteException
     */
    public static int getNewMessageCount(final WebClient webClient) throws WebSiteException { // NOPMD

        int msgCount = 0;

        try {
            final String url = SchwaebischHall.URL_MESSAGE_COUNT;
            final Page responsePage = webClient.getPage(url);
            if (WebUtilsHtmlUnit.webResponseIsJson(responsePage.getWebResponse())) {

                final UnexpectedPage pape = (UnexpectedPage) responsePage;
                Logger.debug("GetNewMessageCount WebResponseCode :" + pape.getWebResponse().getStatusCode()); //$NON-NLS-1$
                Logger.debug(responsePage.getWebResponse().getContentAsString());

                final MessageCountResponse deatilsResponse = BshWebSiteUtils.MAPPER
                        .readValue(responsePage.getWebResponse().getContentAsStream(), MessageCountResponse.class);
                if (deatilsResponse.getStatus() == Status.ERROR) {
                    final List<Error> errors = deatilsResponse.getErrors();
                    for (final Error error : errors) {
                        Logger.error(error.getMessage());
                    }
                    String errorText = BshWebSiteUtils.UNKOWN_ERROR;
                    if (!errors.isEmpty()) {
                        errorText = errors.get(0).getMessage();
                    }
                    throw new WebSiteException(errorText, WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);

                }

                msgCount = deatilsResponse.getMsgCount().getCount();

            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(e.getMessage());
            throw new WebSiteException(I18N.tr("BshWebSiteUtils.getNewMeassagesRequestfailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }

        return msgCount;
    }

    /**
     * Liefert die Anzahl neuer ungelesener Nachrichten.
     *
     * @param webClient
     *            der angemeldete WebClient
     * @return die Anzahl neuer ungelesener Nachrichten
     * @throws WebSiteException
     */
    public static List<MahoeMessage> getNewMessages(final WebClient webClient) throws WebSiteException { // NOPMD

        final List<MahoeMessage> messages = new ArrayList<>();

        try {
            final String url = SchwaebischHall.URL_MESSAGES;
            final Page responsePage = webClient.getPage(url);
            if (WebUtilsHtmlUnit.webResponseIsJson(responsePage.getWebResponse())) {

                final UnexpectedPage pape = (UnexpectedPage) responsePage;
                Logger.debug("GetNewMessages WebResponseCode :" + pape.getWebResponse().getStatusCode()); //$NON-NLS-1$
                Logger.debug(responsePage.getWebResponse().getContentAsString());

                final MessagesResponse detailsResponse = BshWebSiteUtils.MAPPER
                        .readValue(responsePage.getWebResponse().getContentAsStream(), MessagesResponse.class);
                if (detailsResponse.getStatus() == Status.ERROR) {
                    BshWebSiteUtils.logErrors(detailsResponse);

                    throw new WebSiteException(detailsResponse.getErrors().get(0).getMessage(),
                            WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
                }
                BshWebSiteUtils.addNewMessages(detailsResponse, messages);
            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(e.getMessage());
            throw new WebSiteException(I18N.tr("BshWebSiteUtils.getMeassagesRequestFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }

        return messages;
    }

    /**
     * Markiert die Nachricht als gelesen.
     *
     * @param webClient
     * @param message
     * @throws WebSiteException
     */
    public static void markMessageAsRead(final WebClient webClient, final MahoeMessage message)
            throws WebSiteException {

        try {

            final URL url = new URL(SchwaebischHall.URL_POSTBOX);
            final WebRequest requestSetting = new WebRequest(url, HttpMethod.POST);
            requestSetting.setAdditionalHeader(HttpHeaders.USER_AGENT, BshWebSite.CUSTOM_USER_AGENT);
            final List<NameValuePair> urlParameters = new ArrayList<>();
            urlParameters.add(new NameValuePair(BshWebSiteUtils.PARAM_ID, message.getId()));
            urlParameters.add(new NameValuePair(BshWebSiteUtils.PARAM_COMMAND, BshWebSiteUtils.COMMAND_READ));
            requestSetting.setRequestParameters(urlParameters);

            final Page responsePage = webClient.getPage(requestSetting);
            if (WebUtilsHtmlUnit.webResponseIsJson(responsePage.getWebResponse())) {

                final UnexpectedPage pape = (UnexpectedPage) responsePage;
                Logger.debug("GetNewMessages WebResponseCode :" + pape.getWebResponse().getStatusCode()); //$NON-NLS-1$
                Logger.debug(responsePage.getWebResponse().getContentAsString());

                final MessageReadResponse deatilsResponse = BshWebSiteUtils.MAPPER
                        .readValue(responsePage.getWebResponse().getContentAsStream(), MessageReadResponse.class);
                if (deatilsResponse.getStatus() == Status.ERROR) {
                    final List<Error> errors = deatilsResponse.getErrors();
                    for (final Error error : errors) {
                        Logger.error(error.getMessage());
                    }
                    String errorText = BshWebSiteUtils.UNKOWN_ERROR;
                    if (!errors.isEmpty()) {
                        errorText = errors.get(0).getMessage();
                    }
                    throw new WebSiteException(errorText, WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
                }
            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(e.getMessage());
            throw new WebSiteException(I18N.tr("BshWebSiteUtils.markReadRequestFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }
    }

    /**
     * Liefert die IBAN
     *
     * @param webClient
     * @param bsnKurz
     * @param blz
     * @return die IBAN, oder NULL
     * @throws WebSiteException
     */
    /* default */static String getIban(final WebClient webClient, final String bsnKurz, final String blz) // NOPMD
            throws WebSiteException {

        String iban = null;
        try {
            final String url = SchwaebischHall.URL_IBAN + "?bsn=" + bsnKurz + "&blz=" + blz; //$NON-NLS-1$ //$NON-NLS-2$
            final Page responsePage = webClient.getPage(url);
            if (WebUtilsHtmlUnit.webResponseIsJson(responsePage.getWebResponse())) {

                final UnexpectedPage pape = (UnexpectedPage) responsePage;
                Logger.debug("GetIban WebResponseCode :" + pape.getWebResponse().getStatusCode()); //$NON-NLS-1$
                Logger.debug(responsePage.getWebResponse().getContentAsString());

                final IbanResponse ibanResponse = BshWebSiteUtils.MAPPER
                        .readValue(responsePage.getWebResponse().getContentAsStream(), IbanResponse.class);
                if (ibanResponse.getStatus() == Status.ERROR) {
                    final List<Error> errors = ibanResponse.getErrors();
                    for (final Error error : errors) {
                        Logger.error(error.getMessage());
                    }
                    String errorText = BshWebSiteUtils.UNKOWN_ERROR;
                    if (!errors.isEmpty()) {
                        errorText = errors.get(0).getMessage();
                    }
                    throw new WebSiteException(errorText, WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
                }

                iban = ibanResponse.getIban();

            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(e.getMessage());
            throw new WebSiteException(I18N.tr("BshWebSiteUtils.ibanRequestFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }

        return iban;
    }

    /**
     * Liefert einen instanzierten HtmlUnit WebClient
     *
     * @return der WebClient
     */
    /* default */static WebClient getWebClient() { // NOPMD

        final WebClient webcl = new WebClient();

        final WebClientOptions options = webcl.getOptions();

        options.setUseInsecureSSL(false);
        options.setJavaScriptEnabled(true);
        options.setThrowExceptionOnFailingStatusCode(false);
        options.setCssEnabled(true);
        options.setThrowExceptionOnScriptError(false);
        options.setRedirectEnabled(true);

        webcl.setAjaxController(new NicelyResynchronizingAjaxController());
        webcl.getCookieManager().setCookiesEnabled(true);

        Utils.setProxyCfg(webcl, SchwaebischHall.URL_HOME);

        return webcl;

    }

    /**
     * Get all new messages from the response and adds them in the given list
     *
     * @param detailsResponse
     *            the with the messages
     * @param messages
     *            the list to add new messages to
     */
    private static void addNewMessages(final MessagesResponse detailsResponse, final List<MahoeMessage> messages) {

        if (messages != null) {
            final List<Message> msgs = detailsResponse.getMessages();
            for (final Message msg : msgs) {
                if (Boolean.FALSE.equals(msg.isRead())) {
                    String text = msg.getText().replace("\n", StringUtils.EMPTY); //$NON-NLS-1$
                    text = text.replace("<br />", "\n"); //$NON-NLS-1$ //$NON-NLS-2$
                    if (msg.getType().equals("data")) { //$NON-NLS-1$ //
                        // NOPMD
                        final StringBuilder builder = new StringBuilder(text.length() + 106); // NOPMD
                        builder.append(I18N.tr("BshWebSiteUtils.attachmentWarning1") //$NON-NLS-1$
                                + I18N.tr("BshWebSiteUtils.attachmentWarning2")) //$NON-NLS-1$
                                .append(text);
                        text = builder.toString();
                    }

                    final MahoeMessage message = new MahoeMessage(SchwaebischHall.BLZ, msg.getDate(), // NOPMD
                            msg.getTitle(), null);
                    message.setMessage(text);
                    message.setId(msg.getId());
                    messages.add(message);
                }
            }
        }
    }

    /**
     * Check the exemption error for possible occurrence of marriage problem.
     *
     * @param exemptionResponse
     *            the answer to check
     * @throws WebSiteException
     *             in case it was not the marriage problem.
     */
    private static void handleExemptionError(final ExemptionsResponse exemptionResponse) throws WebSiteException {
        final List<Error> errors = exemptionResponse.getErrors();
        for (final Error error : errors) {
            Logger.error(I18N.tr("BshWebSiteUtils.exemptionsError") + error.getMessage()); //$NON-NLS-1$
        }
        if (errors.isEmpty()) {
            throw new WebSiteException(BshWebSiteUtils.UNKOWN_ERROR, WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
        } else {
            final String errorText = errors.get(0).getMessage();
            if (StringUtils.isAllEmpty(errorText) || errorText.startsWith("Freistellungsauftr")) { //$NON-NLS-1$
                // all fine. This is a possible valid error.
                // So nothing to do
                Logger.info(I18N.tr("BshWebSiteUtils.exemptionsErrorMarried")); //$NON-NLS-1$
            } else {
                throw new WebSiteException(errorText, WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
            }
        }

    }

    /**
     * Logs all error from a response
     *
     * @param detailsResponse
     */
    private static void logErrors(final MessagesResponse detailsResponse) {

        final List<Error> errors = detailsResponse.getErrors();
        for (final Error error : errors) {
            Logger.error(error.getMessage());
        }
    }

    /**
     * Privater Konstruktor
     */
    private BshWebSiteUtils() {

        throw new AssertionError("Instantiating utility class..."); //$NON-NLS-1$
    }

}
