/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.pdfextractor;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.text.PDFTextStripper;

import de.frankmuenster.mahoe.banks.santander.clearing.ClearingEntry;
import de.frankmuenster.mahoe.pdfextractor.AbstractBankPDFPage;
import de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor;
import de.frankmuenster.mahoe.pdfextractor.AbstractTransactionCreator;
import de.frankmuenster.mahoe.pdfextractor.PdfBuchung;
import de.frankmuenster.mahoe.pdfextractor.PdfEintrag;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.willuhn.logging.Logger;

/**
 * Class for extracting entry information from a Santanter Bank PDF account
 * statement.
 *
 * @author Frank Mnster
 */
@SuppressWarnings({ "PMD.GodClass", "PMD.TooManyMethods" })
public class SantanderPdfExtractor extends AbstractPdfExtractor {

    /** Der Text, der die letzte Seite mit Umstzen angibt. */
    private static final String LAST_PAGE_TEXT = "NEUER SALDO IN EURO"; //$NON-NLS-1$

    /** Die HABEN Bezeichnung. */
    private static final String HABEN_IDENTIFIER = "H"; //$NON-NLS-1$

    /**
     * Zeilenhhe der Buchungszeile im MahoeMain PDF.
     */
    private static final double LINE_HEIGH = 9.35;

    /**
     * Max. Unterschied ersten Textblocks, um als Buchungsdatum erkannt zu
     * werden.
     */
    private static final double MAX_POS_DATE_DIFF = 0.05;

    /** Dezimalformatierer. */
    private static final NumberFormat N_FORMAT = NumberFormat.getInstance(Locale.GERMAN);

    /** Letzte bei den Fremdwhrungen ausgelese Seite. */
    private PDPage lastPDPage;

    /** Liste mit den Fremdw&auml;hrungen. */
    private List<PdfEintrag> fremdwaehrung;

    /** Liste mit den Fremdw&auml;hrungsbetr&auml;gen. */
    private List<PdfEintrag> fwBetraege;

    /** First page with variable transaction position */
    private SantanderPdfPageFirst pageFirst;

    // TODO: Remove code
    // /** FINANZIERUNGSGEBHR found. Need to add one more line to keep data **/
    // private final Boolean feeFound = false;

    /**
     * Creates a new {@link AbstractPdfExtractor} for the Santander Bank.
     *
     * @param byteArray
     *            the account sheet PDF as byte array
     *
     * @param cardNumber
     *            the credit card number
     * @see AbstractPdfExtractor#AbstractPdfExtractor(byte[], String)
     */
    public SantanderPdfExtractor(final byte[] byteArray, final String cardNumber) {
        super(byteArray, cardNumber);

        try {
            this.pageFirst = new SantanderPdfPageFirst(this.document.getPages().get(0));
        } catch (final IOException e) {
            Logger.error("Fehler beim Auslesen der ersten Seite");
        }

    }

    /**
     * Erstellt einen PDF-Extraktor der Santander Bank.
     *
     * @param inputStream
     *            the input stream
     * @param kartenNummer
     *            the karten nummer
     * @see AbstractPdfExtractor#AbstractPdfExtractor(InputStream, String)
     */
    public SantanderPdfExtractor(final InputStream inputStream, final String kartenNummer) {
        super(inputStream, kartenNummer);

        try {
            this.pageFirst = new SantanderPdfPageFirst(this.document.getPages().get(0));
        } catch (final IOException e) {
            Logger.error("Fehler beim Auslesen der ersten Seite");
        }

    }

    /**
     * Liefert den Ausgleichbetrag. Das ist die Differenz zwischen dem Saldo und
     * dem flligem Betrag.
     *
     * @return den Ausgleichbetrag
     */
    public ClearingEntry getAusgleichsBuchung() {

        ClearingEntry buchung = null;
        final PDPage page = this.document.getPage(0);
        try {
            final AbstractBankPDFPage textAreas = this.pageFirst;
            final int a = 1;
            textAreas.extractRegions(page);
            final String kartenkonto = this.getKartenkonto(textAreas);
            final String karteninhaber = super.getKarteninhaber(textAreas);
            final String iban = super.getIban(textAreas);
            final String bic = super.getBic(textAreas).trim();
            final BigDecimal saldo = this.getSaldo(textAreas);
            final BigDecimal fBetrag = this.getFaelligBetrag(textAreas);
            BigDecimal betrag = BigDecimal.ZERO;
            if (saldo != null && fBetrag != null) {
                betrag = saldo.add(fBetrag).setScale(2, RoundingMode.HALF_UP);
            }
            final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", Locale.GERMAN); //$NON-NLS-1$
            final Date datum = sdf.parse(this.getFaelligDatum(textAreas));
            buchung = new ClearingEntry(kartenkonto, karteninhaber, iban, bic, datum, betrag);
        } catch (IOException | ParseException e) {
            Logger.warn(I18N.tr("SantanderPdfExtractor.extractingDataFailed") + e.getMessage()); //$NON-NLS-1$
            Logger.warn(Arrays.toString(e.getStackTrace()));
        }
        return buchung;
    }

    /**
     * Adds the year.
     *
     * @param entries
     *            the entries
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#addYear
     */
    @Override
    protected void addYear(final List<PdfEintrag> entries) {

        // Jahr hinzu tun
        final int rMonat = Integer.parseInt(this.rechungsDatum.substring(3, 5));
        final int rJahr = Integer.parseInt(this.rechungsDatum.substring(6, 10));
        for (final PdfEintrag entry : entries) {
            final String datum = entry.getWort().substring(0, 2) + "." //$NON-NLS-1$
                    + entry.getWort().substring(2, 4) + "."; //$NON-NLS-1$
            final int bMonat = Integer.parseInt(entry.getWort().substring(2, 4));
            if (bMonat > rMonat) {
                entry.setWort(datum + (rJahr - 1));
            } else {
                entry.setWort(datum + rJahr);
            }
        }
    }

    /**
     * Gets the bank even pdf page.
     *
     * @return the bank even pdf page
     * @throws IOException
     *             Signals that an I/O exception has occurred.
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfPageEven()
     */
    @Override
    protected AbstractBankPDFPage getBankPdfPageEven() throws IOException {

        return new SantanderPdfPageEven();
    }

    /**
     * Gets the bank first pdf page.
     *
     * @return the bank first pdf page
     * @throws IOException
     *             Signals that an I/O exception has occurred.
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfPageFirst()
     */
    @Override
    protected AbstractBankPDFPage getBankPdfPageFirst() throws IOException {

        return this.pageFirst;
    }

    /**
     * Gets the bank odd pdf page.
     *
     * @return the bank odd pdf page
     * @throws IOException
     *             Signals that an I/O exception has occurred.
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfPageOdd()
     */
    @Override
    protected AbstractBankPDFPage getBankPdfPageOdd() throws IOException {

        return new SantanderPdfPageOdd();
    }

    /**
     * Gets the betraege.
     *
     * @param textAreas
     *            the text areas
     * @return the betraege
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getBetraege(de.frankmuenster.mahoe.pdfextractor.AbstractBankPDFPage)
     */
    @Override
    protected List<PdfEintrag> getBetraege(final AbstractBankPDFPage textAreas) {

        final List<PdfEintrag> betraege = new ArrayList<>();
        final String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.BETRAG);
        final String lineSeparator = textAreas.getLineSeparator();
        final String[] lines = regioText.split(lineSeparator);
        for (final String line : lines) {
            Logger.trace("Line (betrag): >" + line + "<"); //$NON-NLS-1$ //$NON-NLS-2$
            if (StringUtils.isNotEmpty(line)) {
                readBetragfromLine(line, betraege);
            }

        }
        // Anmerkungen entfernen
        this.removeAnnotations(betraege);
        return betraege;
    }

    /**
     * Gets the fremdwaehrungen.
     *
     * @param textAreas
     *            the text areas
     * @return the fremdwaehrungen
     * @see AbstractPdfExtractor#getFremdwaehrungen(AbstractBankPDFPage)
     */
    @Override
    protected List<PdfEintrag> getFremdwaehrungen(final AbstractBankPDFPage textAreas) {

        if (!textAreas.getCurrentPage().equals(this.lastPDPage)) {
            this.setFremdWaehrungenAndBetraegeFrom(textAreas);
        }
        return this.fremdwaehrung;
    }

    /**
     * Gets the fremdwaehrungs betraege.
     *
     * @param textAreas
     *            the text areas
     * @return the fremdwaehrungs betraege
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getFremdwaehrungsBetraege(de.frankmuenster.mahoe.pdfextractor.AbstractBankPDFPage)
     */
    @Override
    protected List<PdfEintrag> getFremdwaehrungsBetraege(final AbstractBankPDFPage textAreas) {

        if (!textAreas.getCurrentPage().equals(this.lastPDPage)) {
            this.setFremdWaehrungenAndBetraegeFrom(textAreas);
        }
        return this.fwBetraege;
    }

    /**
     * Gets the last page.
     *
     * @return the last page
     * @throws IOException
     *             Signals that an I/O exception has occurred.
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getLastPage()
     */
    @Override
    protected int getLastPage() throws IOException {

        int lastPage = this.document.getPages().getCount();
        final PDFTextStripper textStripper = new PDFTextStripper();
        for (int pageNumber = 1; pageNumber <= this.document.getPages().getCount(); pageNumber++) {
            textStripper.setStartPage(pageNumber);
            textStripper.setEndPage(pageNumber);
            // Fetch page text and insert into "pages" string
            final String pageText = textStripper.getText(this.document);
            if (pageText.contains(SantanderPdfExtractor.LAST_PAGE_TEXT)) {
                lastPage = pageNumber;
                break;
            }
        }
        return lastPage;
    }

    /**
     * Gets the number format.
     *
     * @return the number format
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getNumberFormat()
     */
    @Override
    protected NumberFormat getNumberFormat() {

        return SantanderPdfExtractor.N_FORMAT;
    }

    /**
     * Gets the saldo.
     *
     * @param textAreas
     *            the text areas
     * @return the saldo
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getSaldo(de.frankmuenster.mahoe.pdfextractor.AbstractBankPDFPage)
     */
    @Override
    protected BigDecimal getSaldo(final AbstractBankPDFPage textAreas) {

        final String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.SALDO);
        final String[] woerter = regioText.split(" "); //$NON-NLS-1$

        final StringBuffer saldoBuffer = new StringBuffer();
        for (final String wort : woerter) {
            final String[] text = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
            if (text.length == 3) {
                saldoBuffer.append(text[2]);
            }
        }

        final String saldoText = saldoBuffer.toString().replace("\n", StringUtils.EMPTY);
        final String[] saldoParts = saldoText.split(" ");

        Logger.trace("PT (saldo): " + saldoParts.length); //$NON-NLS-1$
        String saldoString = "-" + saldoParts[0]; //$NON-NLS-1$
        saldoString = saldoString.replace("\r", StringUtils.EMPTY); //$NON-NLS-1$
        saldoString = saldoString.replace(" ", StringUtils.EMPTY); //$NON-NLS-1$
        BigDecimal saldo = null;
        try {
            saldo = new BigDecimal(this.numFormat.parse(saldoString).toString());
            // Zweites Wort ist H, dann positives Saldo
            if (saldoParts.length == AbstractPdfExtractor.MAX_TWO_WORD
                    && SantanderPdfExtractor.HABEN_IDENTIFIER.equals(saldoParts[2].trim())) {
                saldo = saldo.negate();
            }
        } catch (

        final ParseException e) {
            Logger.warn(I18N.tr("SantanderPdfExtractor.saldoNotFound", e.getMessage())); //$NON-NLS-1$
        }

        return saldo;
    }

    /**
     * Gets the transaction creator.
     *
     * @return the transaction creator
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#getTransactionCreator()
     */
    @Override
    protected AbstractTransactionCreator getTransactionCreator() {

        return new SantanderTransactionCreator(this.buchungsDaten, SantanderPdfExtractor.LINE_HEIGH);
    }

    /**
     * Removes the annotations.
     *
     * @param pdfWortList
     *            the pdf wort list
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#removeAnnotations(java.util.List)
     */
    @Override
    protected void removeAnnotations(final List<PdfEintrag> pdfWortList) {

        final double yOfFirstDatum = this.buchungsDaten.get(this.buchungsDaten.size() - 1).getPositionY() - 10;
        // TODO: Remove
        // if (this.feeFound) {
        // // yOfFirstDatum -= SantanderPdfExtractor.LINE_HEIGH;
        // }
        Logger.trace("First: " + yOfFirstDatum); //$NON-NLS-1$
        for (int i = pdfWortList.size() - 1; i >= 0; i--) {
            Logger.trace("Check: " + pdfWortList.get(i).getPositionY()); //$NON-NLS-1$
            if (pdfWortList.get(i).getPositionY() < yOfFirstDatum) {
                // check if FINANZIERUNGSGEBHR
                Logger.debug(pdfWortList.get(i).getWort());
                if ("FINANZIERUNGSGEBHR".equals(pdfWortList.get(i).getWort())) {
                    // this.feeFound = true;
                    // add dummy dates
                    final PdfEintrag datum = new PdfEintrag(); // NOPMD
                    datum.setPositionX(pdfWortList.get(i).getPositionX());
                    datum.setPositionY(pdfWortList.get(i).getPositionY());
                    datum.setWort(this.rechungsDatum);
                    this.buchungsDaten.add(datum);

                } else {
                    Logger.trace("Remove"); //$NON-NLS-1$
                    pdfWortList.remove(i);
                }
            }
        }
    }

    /**
     * Removes the annotations booking dates.
     *
     * @param buchungenDatum
     *            the buchungen datum
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#removeAnnotationsBookingDates(java.util.List)
     */
    @Override
    protected void removeAnnotationsBookingDates(final List<PdfEintrag> buchungenDatum) {

        if (!buchungenDatum.isEmpty()) {
            final double xOfFirstDatum = buchungenDatum.get(0).getPositionX();
            Logger.trace("First date: " + xOfFirstDatum); //$NON-NLS-1$
            for (int i = buchungenDatum.size() - 1; i > 0; i--) {
                Logger.trace("Check date x position: " + buchungenDatum.get(i).getPositionX()); //$NON-NLS-1$
                final double xPosDiff = Math.abs(buchungenDatum.get(i).getPositionX() - xOfFirstDatum);
                if (xPosDiff > SantanderPdfExtractor.MAX_POS_DATE_DIFF) {
                    Logger.trace("Remove date as it is an annotation in the bottom"); //$NON-NLS-1$
                    buchungenDatum.remove(i);
                }
            }
        }
    }

    /**
     * Re order.
     *
     * @param buchungen
     *            the buchungen
     * @see de.frankmuenster.mahoe.pdfextractor.AbstractPdfExtractor#reOrder(java.util.List)
     */
    @Override
    protected void reOrder(final List<PdfBuchung> buchungen) {

        // Leider mssen wir richtig sortieren, da die Umstze nach Karten
        // sortiert sind
        buchungen.sort((final PdfBuchung buchung1,
                final PdfBuchung buchung2) -> buchung1.getBuchungsDatum().compareTo(buchung2.getBuchungsDatum()) * -1);

        /**
         * Collections.sort(buchungen, new Comparator<PdfBuchung>() {
         *
         * @Override public int compare(final PdfBuchung buchung1, final
         *           PdfBuchung buchung2) {
         *
         *           return
         *           buchung1.getBuchungsDatum().compareTo(buchung2.getBuchungsDatum())
         *           * -1; } });
         */
    }

    /**
     * Adds the currency and value.
     *
     * @param waehrungText
     *            the waehrung text
     * @param betragText
     *            the betrag text
     */
    private void addCurrencyAndValue(final String waehrungText, final String betragText) {
        final PdfEintrag pdfWaehrung = new PdfEintrag(); // NOPMD
        if (StringUtils.isEmpty(waehrungText)) {
            pdfWaehrung.setWort(StringUtils.EMPTY);
        } else {
            final String[] wortInfo = waehrungText.split(AbstractPdfExtractor.WORT_DELIMITER);
            pdfWaehrung.setPositionX(Double.valueOf(wortInfo[0]));
            pdfWaehrung.setPositionY(Double.valueOf(wortInfo[1]));
            pdfWaehrung.setWort(Utils.removeEOL(wortInfo[2]));
        }
        final PdfEintrag pdfBetrag = new PdfEintrag(); // NOPMD
        if (StringUtils.isEmpty(betragText)) {
            pdfBetrag.setWort(StringUtils.EMPTY);
        } else {
            final String[] wortInfo = betragText.split(AbstractPdfExtractor.WORT_DELIMITER);
            pdfBetrag.setPositionX(Double.valueOf(wortInfo[0]));
            pdfBetrag.setPositionY(pdfWaehrung.getPositionY());
            pdfBetrag.setWort(Utils.removeEOL(wortInfo[2]));
        }
        this.fremdwaehrung.add(pdfWaehrung);
        this.fwBetraege.add(pdfBetrag);

    }

    /**
     * Gets the faellig betrag.
     *
     * @param textAreas
     *            the text areas
     * @return the faellig betrag
     */
    private BigDecimal getFaelligBetrag(final AbstractBankPDFPage textAreas) {

        final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.FAELLIG_BETRAG)
                .split(AbstractPdfExtractor.WORT_DELIMITER);

        String fBetragString;

        if (parsedText.length > 2) {
            // text is split into multiple parts :-(
            final StringBuilder tmpBuf = new StringBuilder();

            for (int i = 2; i < parsedText.length - 1; i = i + 2) {
                final String temp[] = parsedText[i].split(" ");
                tmpBuf.append(temp[0]);
            }
            tmpBuf.append(parsedText[parsedText.length - 1]);

            fBetragString = tmpBuf.toString();
        } else {
            fBetragString = Utils.removeEOL(parsedText[2]);
        }

        fBetragString = fBetragString.replace(" ", StringUtils.EMPTY); //$NON-NLS-1$
        BigDecimal fBetrag = null;
        try {
            fBetrag = BigDecimal.valueOf(this.numFormat.parse(fBetragString).doubleValue());
        } catch (final ParseException e) {
            Logger.warn(I18N.tr("SantanderPdfExtractor.clearingAmountNotFound") + e.getMessage()); //$NON-NLS-1$
        }
        return fBetrag;
    }

    /**
     * Liefert das Flligkeitsdatum des Saldos.
     *
     * @param textAreas
     *            die Textares der ersten Seite
     * @return das Flligkeitsdatum
     */
    private String getFaelligDatum(final AbstractBankPDFPage textAreas) {

        final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.FAELLIG_DATUM)
                .split(AbstractPdfExtractor.WORT_DELIMITER);

        String datum;
        if (parsedText.length > 2) {
            // text is split into multible parts :-(
            Logger.debug("PT (rechnungsdatum): " + parsedText.length);
            final StringBuilder datumTmp = new StringBuilder();

            for (int i = 2; i < parsedText.length - 1; i = i + 2) {
                final String temp[] = parsedText[i].split(" ");
                datumTmp.append(temp[0]);
            }
            datumTmp.append(parsedText[parsedText.length - 1]);

            datum = datumTmp.toString();
        } else {

            datum = parsedText[2];
        }
        datum = datum.replace("\n", "").replace("\r", "");

        return Utils.removeEOL(datum);
    }

    /**
     * Read the value of the transaction of one line and adds it to the list of
     * values.
     *
     * @param line
     *            the line to inspect
     * @param betraege
     *            the list to add the found value too
     */
    private void readBetragfromLine(final String line, final List<PdfEintrag> betraege) {

        final String[] woerter = line.split(" "); //$NON-NLS-1$
        if (woerter.length > 0 && woerter.length <= AbstractPdfExtractor.MAX_TWO_WORD) {

            boolean isHaben = false;
            if (woerter.length == AbstractPdfExtractor.MAX_TWO_WORD) {
                // Wenn das zweite Wort nur ein H ist, haben wir eine
                // Einzahlung
                final String[] testHaben = woerter[1].split(AbstractPdfExtractor.WORT_DELIMITER);
                if (SantanderPdfExtractor.HABEN_IDENTIFIER.equals(testHaben[2].trim())) {
                    isHaben = true;
                }
            }
            final String wort = woerter[0];
            final String[] wortInfo = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
            if (wortInfo.length == AbstractPdfExtractor.WORT_INFO_LENGTH) {
                if (!isHaben) {
                    wortInfo[2] = '-' + wortInfo[2]; // NOPMD
                }
                final PdfEintrag pdfEintrag = new PdfEintrag(); // NOPMD
                pdfEintrag.setPositionX(Double.valueOf(wortInfo[0]));
                pdfEintrag.setPositionY(Double.valueOf(wortInfo[1]));
                pdfEintrag.setWort(Utils.removeEOL(wortInfo[2]));
                betraege.add(pdfEintrag);
            }
        }
    }

    /**
     * Sets the fremd waehrungen and betraege from.
     *
     * @param textAreas
     *            the text area to take the currency and value from
     */
    private void setFremdWaehrungenAndBetraegeFrom(final AbstractBankPDFPage textAreas) {

        this.fremdwaehrung = new ArrayList<>();
        this.fwBetraege = new ArrayList<>();
        String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.FREMDW_BETRAG);
        final String lineSeparator = textAreas.getLineSeparator();
        if (lineSeparator.equals(regioText)) {
            regioText = StringUtils.EMPTY;
        }
        final String[] lines = regioText.split(lineSeparator);
        boolean multiLine = false;
        String betragText;
        String waehrungText = StringUtils.EMPTY;
        // Loop over all liens
        for (final String line : lines) {
            Logger.debug("Line (fwBetrag): >" + line + "<"); //$NON-NLS-1$//$NON-NLS-2$
            if (StringUtils.isNotEmpty(line)) {
                final String[] woerter = line.split(" "); //$NON-NLS-1$
                Logger.debug("Anzahl Wrter: " + woerter.length); //$NON-NLS-1$
                if (woerter.length == AbstractPdfExtractor.MAX_ONE_WORD) {
                    if (multiLine) {
                        multiLine = false;
                        betragText = woerter[0];
                        // Now add currency and value to lists if they exists
                        addCurrencyAndValue(waehrungText, betragText);
                    } else {
                        multiLine = true;
                        waehrungText = woerter[0];
                    }
                } else if (woerter.length == AbstractPdfExtractor.MAX_TWO_WORD) {
                    multiLine = false;
                    waehrungText = woerter[0];
                    betragText = woerter[1];
                    // Now add currency and value to lists if they exists
                    addCurrencyAndValue(waehrungText, betragText);
                }
            }
        }
        // Anmerkungen entfernen
        this.removeAnnotations(this.fremdwaehrung);
        this.removeAnnotations(this.fwBetraege);
        // Seite speichern, damit alte Werte nicht wieder verwendet werden
        this.lastPDPage = textAreas.getCurrentPage();
    }

}
