/*******************************************************************************
 * Copyright 2015 - 2019 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.webpages;

import java.math.BigDecimal;
import java.time.Duration;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.openqa.selenium.By;
import org.openqa.selenium.TimeoutException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedCondition;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import de.frankmuenster.mahoe.banks.santander.web.SantanderWebAccountInformation;
import de.frankmuenster.mahoe.banks.santander.web.SantanderWebSite;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.frankmuenster.mahoe.web.webpages.DetailsPage;
import de.willuhn.logging.Logger;

/**
 * Page with account details
 *
 * @author Frank Mnster
 * @since 2.4.0
 *
 */
@SuppressWarnings({ "PMD.TooManyMethods" })
public class SantanderDetailsPage extends SantanderTransactionsPage implements DetailsPage {

    /** I18n key for error parsing a date */
    private static final String I18N_ERROR_PARSING_DATE = "DetailsPage.errorParsingDate"; //$NON-NLS-1$

    /** the saldo */
    private BigDecimal saldo;

    /** credit Limit */
    private BigDecimal creditLimit;

    /** date of the last transaction */
    private LocalDate lastTransaction;

    /** date of next subtraction */
    private LocalDate nextSubtraction;

    /** date of next bill */
    private LocalDate nextBill;

    /** the available amount */
    private BigDecimal availableAmount;

    /** the product description */
    private String productDescription;

    /**
     * Account to open the details page for
     */

    /**
     * Create a new instance of the details page for the given account
     *
     * @param driver
     * @param accountInfo
     * @throws WebSiteException
     */
    public SantanderDetailsPage(final WebDriver driver, final String baseUrl,
            final SantanderWebAccountInformation accountInfo) throws WebSiteException {
        // open the account page
        super(driver, baseUrl, accountInfo);

        // get Saldo
        setSaldo();

        // if credit card get other details
        if (super.isCreditCard) {
            setAvailable();
            openDetails();
            setDetails();
            closeDetails();
        }

    }

    /**
     * Returns the available amount.
     *
     * @return the available amount
     * @throws WebSiteException
     *             in case the value could not be read from the details hover
     *             page
     */
    @Override
    public BigDecimal getAvailableAmount() throws WebSiteException {
        if (this.availableAmount == null) {
            throw new WebSiteException(I18N.tr("DetailsPage.availableAmountNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
        }
        return this.availableAmount;
    }

    /**
     * Return the limit of the credit card.
     *
     * @return the credit card limit
     * @throws WebSiteException
     *             in case the value could not be read from the details hover
     *             page
     */
    @Override
    public BigDecimal getCreditLimit() throws WebSiteException {
        if (this.creditLimit == null) {
            throw new WebSiteException(I18N.tr("DetailsPage.limitNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
        }
        return this.creditLimit;
    }

    /**
     * Return the date of the last transaction.
     *
     * @return date of the last transaction
     * @throws WebSiteException
     *             in case the value could not be read from the details hover
     *             page
     */
    @Override
    public LocalDate getLastTransaction() throws WebSiteException {
        if (this.lastTransaction == null) {
            throw new WebSiteException(I18N.tr("DetailsPage.lastTransactionDateNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
        }
        return this.lastTransaction;
    }

    /**
     * Return the date of the next credit card bill.
     *
     * @return date of the next credit card bill
     * @throws WebSiteException
     *             in case the value could not be read from the details hover
     *             page
     */
    @Override
    public LocalDate getNextBill() throws WebSiteException {
        if (this.nextBill == null) {
            throw new WebSiteException(I18N.tr("DetailsPage.nextBillDateNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
        }
        return this.nextBill;
    }

    /**
     * Return the date of the next automatic subtraction.
     *
     * @return date of the next automatic subtraction
     * @throws WebSiteException
     *             in case the value could not be read from the details hover
     *             page
     */
    @Override
    public LocalDate getNextSubtraction() throws WebSiteException {
        if (this.nextSubtraction == null) {
            throw new WebSiteException(I18N.tr("DetailsPage.nextSubstractionNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
        }
        return this.nextSubtraction;
    }

    /**
     * The description of the credit card.
     *
     * @return the credit card description
     * @throws WebSiteException
     *             in case the value could not be read from the details hover
     *             page
     */
    @Override
    public String getProductDescription() throws WebSiteException {
        if (this.productDescription == null) {
            throw new WebSiteException(I18N.tr("DetailsPage.descriptionNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
        }
        return this.productDescription;
    }

    /**
     * The actual saldo.
     *
     * @return the saldo
     * @throws WebSiteException
     *             in case the value could not be read from the details hover
     *             page
     */
    @Override
    public BigDecimal getSaldo() throws WebSiteException {
        if (this.saldo == null) {
            throw new WebSiteException(I18N.tr("DetailsPage.saldoNotFound"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED);
        }
        return this.saldo;
    }

    /**
     * Close the details hover section
     */
    private void closeDetails() {

        // wait for the overlay to be removed
        SantanderWebSite.waitForLoadingOverlay(this.driver);

        // get a wait instance
        final WebDriverWait wait = new WebDriverWait(super.driver, Duration.ofSeconds(10));

        try {
            // find the close button
            // final WebElement closeBtn =
            // super.driver.findElement(By.className("accountDetail__btnClose___I07t4"));
            // //$NON-NLS-1$
            final WebElement closeBtn = super.driver
                    .findElement(By.cssSelector("button[class^=CardContractDetailsContent__btnClose]")); //$NON-NLS-1$
            closeBtn.click();
            wait.until(ExpectedConditions.visibilityOfElementLocated(
                    By.cssSelector("div[class^=SBHeaderMenuLinkLogout__SBHeaderMenuLinkLogoutContainer]"))); //$NON-NLS-1$
        } catch (final TimeoutException e) {
            // only log
            Logger.error(I18N.tr("DetailsPage.closeBtnNotFound"), e); //$NON-NLS-1$
        }

    }

    /**
     * Open the details hove section
     */
    private void openDetails() {

        // wait for loading is finished
        SantanderWebSite.waitForLoadingOverlay(this.driver);

        // get the wait instance
        final WebDriverWait wait = new WebDriverWait(super.driver, Duration.ofSeconds(10));

        // get the details button
        WebElement detailsBtn;
        if (this.isCreditCard) {
            detailsBtn = super.driver.findElement(By.cssSelector("button[class^=CardsMovementsHeader__kartenInfoBtn]")); //$NON-NLS-1$
        } else {
            detailsBtn = super.driver.findElement(By.cssSelector("button[class^=movementsHeader__detailsButton")); //$NON-NLS-1$
        }
        // wait until button is visible and clickable
        wait.until(ExpectedConditions.and(ExpectedConditions.visibilityOf(detailsBtn),
                ExpectedConditions.elementToBeClickable(detailsBtn)));
        // now click
        detailsBtn.click();

        // wait for loading is finished
        SantanderWebSite.waitForLoadingOverlay(this.driver);

        //// wait until the details title is visible
        wait.until(ExpectedConditions.or(
                ExpectedConditions
                        .visibilityOfElementLocated(By.className("accountDetail__accountDetailsTitle___HelZ0")), //$NON-NLS-1$
                ExpectedConditions.visibilityOfElementLocated(
                        By.cssSelector("h2[class^=CardContractDetailsContent__accountDetailsTitle")) //$NON-NLS-1$

        ));

    }

    /**
     * Sets the available amount
     */
    private void setAvailable() {

        final WebElement amountElement = super.driver
                .findElement(By.cssSelector("p[class^=CardsMovementsHeader__amountAvailableDesktop")); //$NON-NLS-1$

        (new WebDriverWait(this.driver, Duration.ofSeconds(10))).until(new ExpectedCondition<Boolean>() {
            @Override
            public Boolean apply(final WebDriver driver) {
                return amountElement.getText().length() > 1;
            }
        });

        String amount = amountElement.getText();
        Logger.debug("Verfgbar: " + amount);
        amount = amount.substring(0, amount.length() - 2);

        this.availableAmount = Utils.convertStringToBigDecimal(amount);

    }

    /**
     * Sets the details
     */
    private void setDetails() {

        final DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd.MM.yyyy", Locale.GERMAN); //$NON-NLS-1$

        final Map<String, String> detailsMap = new ConcurrentHashMap<>();

        final WebElement detailsList = super.driver
                .findElement(By.className("CardContractDetailsContent__listContainer___N2LAf")); //$NON-NLS-1$
        final String details = detailsList.getText();
        Logger.debug("Details: " + details); //$NON-NLS-1$

        final String[] detailsAry = details.split("\n"); //$NON-NLS-1$

        for (int i = 0; i < detailsAry.length; i = i + 2) {
            detailsMap.put(detailsAry[i], detailsAry[i + 1]);
        }

        // dispo credit
        String detailText = detailsMap.get("Limit"); //$NON-NLS-1$
        detailText = detailText.substring(0, detailText.length() - 2);
        if (detailText != null) {
            this.creditLimit = Utils.convertStringToBigDecimal(detailText);
        }

        // Last transaction Letzte Abrechnung
        detailText = detailsMap.get("Letzte Abrechnung"); //$NON-NLS-1$
        if (detailText != null) {
            try {
                this.lastTransaction = LocalDate.parse(detailText, formatter);
            } catch (final DateTimeParseException e) {
                Logger.error(I18N.tr(SantanderDetailsPage.I18N_ERROR_PARSING_DATE, detailText, "Letzte Abrechnung"), e); //$NON-NLS-1$
            }
        }

        // next Flligkeit - Flligkeit nchste Lastschrift
        detailText = detailsMap.get("Flligkeit nchste Lastschrift"); //$NON-NLS-1$
        if (detailText != null) {
            try {
                this.nextSubtraction = LocalDate.parse(detailText, formatter);
            } catch (final DateTimeParseException e) {
                Logger.error(
                        I18N.tr(SantanderDetailsPage.I18N_ERROR_PARSING_DATE, detailText,
                                "Flligkeit nchste Lastschrift"), //$NON-NLS-1$
                        e);
            }
        }

        // next bill Nchste Kreditkartenabrechnung
        detailText = detailsMap.get("Nchste Kartenabrechnung"); //$NON-NLS-1$
        if (detailText != null) {
            try {
                this.nextBill = LocalDate.parse(detailText, formatter);
            } catch (final DateTimeParseException e) {
                Logger.error(
                        I18N.tr(SantanderDetailsPage.I18N_ERROR_PARSING_DATE, detailText, "Nchste Kartenabrechnung"), //$NON-NLS-1$
                        e);
            }
        }

        // product description
        detailText = detailsMap.get("Kontoname"); //$NON-NLS-1$
        if (detailText != null) {
            this.productDescription = detailText;
        }

    }

    /**
     * Set the actual saldo
     */
    private void setSaldo() {

        WebElement amountElement;
        if (this.isCreditCard) {
            amountElement = super.driver.findElement(By.cssSelector("span[class^=CardsMovementsHeader__kartenAmount]")); //$NON-NLS-1$

        } else {
            amountElement = super.driver.findElement(By.cssSelector("div[class^=OverallSaldoMH__amount")); //$NON-NLS-1$

        }
        String amount = amountElement.getText();
        Logger.debug("Aktuelles Saldo: " + amount);
        amount = amount.substring(0, amount.length() - 2);

        this.saldo = Utils.convertStringToBigDecimal(amount);
    }

}
