
package de.frankmuenster.mahoe.banks.targobank.web.webpages;

import java.text.ParseException;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.openqa.selenium.By;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import de.frankmuenster.mahoe.banks.targobank.TargoBank;
import de.frankmuenster.mahoe.banks.targobank.web.TargoBankWebAccountInformation;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.utils.WebUtilsSelenium;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.frankmuenster.mahoe.web.webpages.AbstractPostBoxPage;
import de.frankmuenster.mahoe.web.webpages.PostBoxPage;
import de.willuhn.logging.Logger;

/**
 * The post box page for the Targo Bank providing access to Account sheets and
 * other mail/documents
 *
 * @author Frank Mnster
 * @since 2.4.0
 *
 */
public final class TargoBankPostBoxPage extends AbstractPostBoxPage<TargoBankWebAccountInformation, String>
        implements PostBoxPage<String> {

    /** private instance */
    private static TargoBankPostBoxPage instance;

    /**
     * Removes the instance, to avoid old link data
     */

    @SuppressWarnings("PMD.NullAssignment")
    public static void cleanInstance() {
        if (TargoBankPostBoxPage.instance != null) {
            TargoBankPostBoxPage.instance = null;
        }

    }

    /**
     * Instantiate a new PostBoxPage in the given Webdriver an for the given
     * account
     *
     * @param driver
     * @param accountInfo
     *
     * @throws WebSiteException
     */
    public static synchronized TargoBankPostBoxPage getInstance(final WebDriver driver, // NOPMD
                                                                                        // on
                                                                                        // 26.04.22,
                                                                                        // 17:10
            final TargoBankWebAccountInformation accountInfo) throws WebSiteException {
        if (TargoBankPostBoxPage.instance == null) {
            TargoBankPostBoxPage.instance = new TargoBankPostBoxPage(driver, accountInfo);
        } else {
            // check accountInfo are identical
            if (!accountInfo.getCardNumber().equals(TargoBankPostBoxPage.instance.accountInfo.getCardNumber())) {
                TargoBankPostBoxPage.instance = new TargoBankPostBoxPage(driver, accountInfo);
            }
        }
        return TargoBankPostBoxPage.instance;
    }

    /**
     * Instantiate a new PostBoxPage in the given Webdriver an for the given
     * account
     *
     * @param driver
     * @param accountInfo
     *
     * @throws WebSiteException
     */
    private TargoBankPostBoxPage(final WebDriver driver, final TargoBankWebAccountInformation accountInfo)
            throws WebSiteException {

        super(driver, accountInfo);

        // Initialise lists
        this.accountSheetDocuments = new ArrayList<>();
        this.otherDocuments = new ArrayList<>();

        // Initialise list positions
        this.accountSheetsPosition = 0;
        this.documentsPosition = 0;

        openPage();
        // Load all documents in a list
        loadDocuments();
        // load additional correspondents
        loadAdditioanCorrespondents();
        // load contact files
        loadContractFiles();

    }

    /**
     * Load additional correspondents files
     *
     * @throws WebSiteException
     */
    private void loadAdditioanCorrespondents() throws WebSiteException {

        // get links to further documents
        final List<WebElement> links = this.driver.findElements(By.id("C:F2_1.T3:li")); //$NON-NLS-1$
        Logger.debug(links.get(0).getText());
        if (!links.isEmpty()) {
            links.get(0).click();

            // wait for loading to be finished
            final WebDriverWait waitLoading = new WebDriverWait(this.driver, Duration.ofSeconds(20));
            waitLoading.until(
                    ExpectedConditions.textToBePresentInElementLocated(By.id("C:S1:F11:T:title"), "Korrespondenz")); //$NON-NLS-1$
        }

        loadDocumentsFromTable(this.otherDocuments, null);

    }

    /**
     * Load additional contract files
     *
     * @throws WebSiteException
     */
    private void loadContractFiles() throws WebSiteException {
        // get links to further documents
        final List<WebElement> links = this.driver.findElements(By.id("C:F2_2.T3:N")); //$NON-NLS-1$

        if (links.get(0).getText().contains("Vertragsunterlagen")) { //$NON-NLS-1$
            links.get(0).click();

            // wait for loading to be finished
            final WebDriverWait waitLoading = new WebDriverWait(this.driver, Duration.ofSeconds(20));
            waitLoading.until(ExpectedConditions.textToBePresentInElementLocated(By.id("C:S1:F11:T:title"), //$NON-NLS-1$
                    "Vertragsunterlagen")); //$NON-NLS-1$
        }

        loadDocumentsFromTable(this.otherDocuments, null);

    }

    /**
     * Currently only loads all transaction sheets
     *
     * @throws WebSiteException
     */
    private void loadDocuments() throws WebSiteException {

        final String number = this.accountInfo.getCardNumber().replace(" ", StringUtils.EMPTY); //$NON-NLS-1$
        boolean found = false;

        // select only the correct card
        final WebElement documentFor = this.driver.findElement(By.id("C:S1:F2_0.dropDownCritSec:DataEntry")); //$NON-NLS-1$

        final List<WebElement> options = documentFor.findElements(By.tagName("option")); //$NON-NLS-1$
        for (final WebElement option : options) {
            Logger.debug(String.format("DocumentFor: %s (CardNumber: %s)", option.getText(), number)); //$NON-NLS-1$
            if (option.getText().endsWith(number)) {
                option.click();
                found = true;
                break;
            }
        }

        if (found) {
            final WebElement search = this.driver.findElement(By.id("C:S1:R4:link")); //$NON-NLS-1$
            search.click();

            // wait for loading to be finished
            final WebDriverWait waitLoading = new WebDriverWait(this.driver, Duration.ofSeconds(20));
            waitLoading.until(ExpectedConditions.invisibilityOfElementLocated(By.xpath(
                    "//img[@src='https://cdnii.e-i.com/INGR/sd/targobank_de_2019/0.80.12/de/images/std/load_ajax.gif']"))); //$NON-NLS-1$

            final String test1 = this.driver.getPageSource();
            Logger.debug(test1);

            // load documents into accountSheet list
            loadDocumentsFromTable(this.accountSheetDocuments, "Ihre Kreditkartenabrechnung");

        }
    }

    private void loadDocumentsFromTable(final List<Pair<Date, String>> list, final String testString)
            throws WebSiteException {

        final WebElement table = this.driver.findElement(By.id("C:S1:panelListeDocs.listeDocs:T")); //$NON-NLS-1$
        final WebElement tbBody = table.findElement(By.tagName(WebUtils.HTML_TAG_TBODY));
        final List<WebElement> rows = tbBody.findElements(By.tagName(WebUtils.HTML_TAG_TR));
        for (final WebElement row : rows) {

            // scroll into view
            WebUtilsSelenium.scrollTo(this.driver, row);
            Logger.debug(String.format("Row: %s", row.getText())); //$NON-NLS-1$

            final String dateTxt = row.findElement(By.xpath("td[2]")).getText(); //$NON-NLS-1$
            Logger.debug(String.format("Date: %s", dateTxt)); //$NON-NLS-1$
            Date date;
            try {
                date = Utils.convertTextToDate(dateTxt, "dd.MM.yyyy"); //$NON-NLS-1$
            } catch (final ParseException e) {
                throw new WebSiteException("Fehler beim Parsen des Datums",
                        WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
            }

            final WebElement link = row.findElement(By.xpath("td[3]")); //$NON-NLS-1$
            // extract hyperlink
            final String txtLink = link.getAttribute("innerHTML");

            if (testString == null || (testString != null && link.getText().contains(testString))) { // NOPMD
                // create the pair and add document to list
                final Pair<Date, String> document = Pair.of(date, txtLink);
                list.add(document);

            } else {
                Logger.info(String.format("Etwas anders als eine Kredikartenabrechnung wurde gefunden\n%s",
                        link.getText()));
            }
        }

    }

    private void openPage() {

        this.driver.get(TargoBank.URL_MESSAGES);

        try {
            final WebElement otherDocuments = this.driver.findElement(By.id("C:R2:N")); //$NON-NLS-1$

            // open all documents
            otherDocuments.click();

        } catch (final org.openqa.selenium.NoSuchElementException e) {
            Logger.error("Other documents link not found.", e);
        }
    }

}
