/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.utils;

import java.math.BigDecimal;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import de.frankmuenster.mahoe.pdfextractor.PdfBuchung;
import de.willuhn.datasource.rmi.DBIterator;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.messaging.ImportMessage;
import de.willuhn.jameica.hbci.messaging.SaldoMessage;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Speichert die Umstze im Hibiscus Konto
 *
 * @author Frank Mnster
 */
@SuppressWarnings({ "PMD.TooManyMethods" })
public final class TransactionSaver {

    /**
     * META Attribute, das einen zu l&ouml;schenden Umsatz markiert.
     */
    private static final String DELETABLE = "PreBookToDelete"; //$NON-NLS-1$

    /**
     * META Attribute Wert, der einen zu l&ouml;schenden Umsatz markiert.
     */
    private static final String DELETABLE_TRUE = "true"; //$NON-NLS-1$

    /**
     * Liefert eine List von Hibiscus Ums&auml;tzen aus einer Liste von
     * PDFBuchungen.
     *
     * @param pdfTransactions
     * @return Liste mit Hibiscus Ums&auml;tzen
     * @throws RemoteException
     */
    public static List<Umsatz> getTransactions(final List<PdfBuchung> pdfTransactions) throws RemoteException {

        final HBCIDBService dbService = Settings.getDBService();
        final List<Umsatz> transactions = new ArrayList<>();

        final StringBuilder foreignInfo = new StringBuilder();
        for (final PdfBuchung pdfTransaction : pdfTransactions) {
            final Umsatz transaction = dbService.createObject(Umsatz.class, null);
            transaction.setDatum(pdfTransaction.getKaufDatum());
            transaction.setValuta(pdfTransaction.getBuchungsDatum());
            transaction.setBetrag(pdfTransaction.getBetrag().doubleValue());
            Utils.loadUmsatz(transaction, pdfTransaction.getZweck());
            // Fremdwhrung
            if (pdfTransaction.getFremdWaehrungsBetrag() != null) {
                foreignInfo.setLength(0);
                foreignInfo.append(I18N.tr("TransactionSaver.amount")) //$NON-NLS-1$
                        .append(pdfTransaction.getFremdWaehrungsBetrag()).append(' ')
                        .append(pdfTransaction.getFremdWaehrung()).append('\n')
                        .append(I18N.tr("TransactionSaver.exchangeRate")) //$NON-NLS-1$
                        .append(pdfTransaction.getKurs());
                transaction.setKommentar(foreignInfo.toString());
            }

            transaction.setSaldo(pdfTransaction.getSaldo().doubleValue());
            transactions.add(transaction);
        }

        return transactions;
    }

    /**
     * Speichert die Liste der Ums&auml;tze im Konto. Die Ums&auml;tze werden
     * vor dem Speichern sortiert.
     *
     * @param account
     *            das Konto
     * @param booked
     *            die Umstze
     * @return Die Anzahl an gespeicherten Ums&auml;tzen
     */
    public static int saveBooked(final Konto account, final List<Umsatz> booked) {

        return TransactionSaver.saveBooked(account, booked, true);
    }

    /**
     * Speichert die Liste der Ums&auml;tze im Konto
     *
     * @param account
     *            das Konto
     * @param booked
     *            die Umstze
     * @param sort
     *            sollen die Umstze vorher nach Datum sortiert werden.
     * @return Die Anzahl an gespeicherten Ums&auml;tzen
     */
    public static int saveBooked(final Konto account, final List<Umsatz> booked, final boolean sort) {

        int savedTransactions = 0;

        if (sort) {
            TransactionSaver.orderTransactions(booked);
        }

        for (final Umsatz umsatz : booked) {
            try {
                umsatz.setKonto(account);
                if (TransactionSaver.isDoubleEntry(account, umsatz)) {
                    Logger.info(I18N.tr("TransactionSaver.transactionExists")); //$NON-NLS-1$
                } else {
                    Logger.debug(I18N.tr("TransactionSaver.saveTransactionLogDebug", //$NON-NLS-1$
                            umsatz.getDatum().toString(), Double.toString(umsatz.getBetrag()),
                            Double.toString(umsatz.getSaldo())));
                    umsatz.store();
                    savedTransactions++;
                    // Nachricht an Hibiscus schicken
                    Application.getMessagingFactory().sendMessage(new ImportMessage(umsatz)); // NOPMD
                }
            } catch (RemoteException | ApplicationException e) {
                Logger.error(I18N.tr("TransactionSaver.addTransactionFailed") + e.getMessage()); //$NON-NLS-1$
            }
        }
        return savedTransactions;
    }

    /**
     * Speichert die vorgemerkten Ums&auml;tze im Konto
     *
     * @param account
     *            das Konto
     * @param preBooked
     *            die Liste der vorgemerkten Ums&auml;tze
     * @return Die Anzahl der gespeicherten Ums&auml;tzen
     */
    public static int savePreBooked(final Konto account, final List<Umsatz> preBooked) {

        return TransactionSaver.savePreBooked(account, preBooked, true);
    }

    /**
     * Speichert die vorgemerkten Ums&auml;tze im Konto
     *
     * @param account
     *            das Konto
     * @param preBooked
     *            die Liste der vorgemerkten Ums&auml;tze
     * @param sort
     *            sollen die Umstze vorher nach Datum sortiert werden.
     * @return Die Anzahl der gespeicherten Ums&auml;tzen
     */
    public static int savePreBooked(final Konto account, final List<Umsatz> preBooked, final boolean sort) {

        // Erst alle vorgemerkten Umstze markieren
        TransactionSaver.markPreBooked(account);
        // Neue speichern bzw. Markierung wieder entfernen
        final int savedTransactions = TransactionSaver.saveBooked(account, preBooked, sort);
        // Nun die alten lschen
        TransactionSaver.removeNotBooked(account);
        return savedTransactions;
    }

    /**
     * Speichert das Saldo im Konto
     *
     * @param konto
     *            das Konto
     * @param saldo
     *            das Saldo
     * @throws ApplicationException
     */
    public static void saveSaldo(final Konto konto, final BigDecimal saldo) throws ApplicationException {

        TransactionSaver.saveSaldo(konto, saldo.doubleValue());
    }

    /**
     * Speichert das Saldo im Konto
     *
     * @param konto
     *            das Konto
     * @param saldo
     *            das Saldo
     * @throws ApplicationException
     */
    public static void saveSaldo(final Konto konto, final Double saldo) throws ApplicationException {

        try {
            konto.setSaldo(saldo);
            konto.store();
        } catch (final RemoteException e) {
            Logger.error(I18N.tr("TransactionSaver.setSaldoFailed"), e); //$NON-NLS-1$
            throw new ApplicationException(I18N.tr("TransactionSaver.setSaldoFailed"), e); //$NON-NLS-1$
        }
        Application.getMessagingFactory().sendMessage(new SaldoMessage(konto));

    }

    /**
     * Speichert den verfgbaren Betrag im Konto
     *
     * @param konto
     *            das Konto
     * @param available
     *            der verfgbaren Betrag
     * @throws ApplicationException
     */
    public static void saveSaldoAvailable(final Konto konto, final BigDecimal available) throws ApplicationException {

        try {
            konto.setSaldoAvailable(available.doubleValue());
            konto.store();
        } catch (final RemoteException e) {
            Logger.error(I18N.tr("TransactionSaver.setSaldoAvailableFailed"), e); //$NON-NLS-1$
            throw new ApplicationException(I18N.tr("TransactionSaver.setSaldoAvailableFailed"), e); //$NON-NLS-1$
        }

    }

    /**
     * berprft, ob ein Umsatz schon im Konto vorhanden ist. Als
     * schlussendliches Merkmal wird noch das Saldo hinzu gezogen.
     *
     * @param account
     *            das Konto
     * @param transaction
     *            der Umsatz
     * @return true wenn der Umsatz schon existiert, false sonst.
     */
    private static boolean isDoubleEntry(final Konto account, final Umsatz transaction) {

        boolean equal = false;
        try {
            final Date umsatzDate = transaction.getDatum();
            final DBIterator<?> umsaetze = account.getUmsaetze(umsatzDate, umsatzDate);
            Logger.debug("DoubleTest -Umsaetze: " + umsaetze.size()); //$NON-NLS-1$
            while (umsaetze.hasNext()) {
                final Umsatz umsatzKonto = (Umsatz) umsaetze.next();
                if (umsatzKonto.equals(transaction) && umsatzKonto.getSaldo() == transaction.getSaldo()) {
                    // Umstze identisch (zu 99,999999%)
                    equal = true;
                    // Ggf. PreBookToDelete entfernen
                    if (TransactionSaver.DELETABLE_TRUE
                            .equals(umsatzKonto.getMeta(TransactionSaver.DELETABLE, StringUtils.EMPTY))) {
                        umsatzKonto.setMeta(TransactionSaver.DELETABLE, StringUtils.EMPTY);
                    }
                    break;
                }

            }
        } catch (final RemoteException e) {
            Logger.warn(I18N.tr("TransactionSaver.isDoubleCheckFailed") + e.getMessage()); //$NON-NLS-1$
        }
        return equal;
    }

    /**
     * Markiert vorhandene vorgemerkten Ums&auml;tze eines Konto.
     *
     * @param account
     *            aus dem alle vorgemerkten Ums&auml;ttze markiert werden
     *            sollen.
     */
    private static void markPreBooked(final Konto account) {

        try {
            final DBIterator<?> umsaetze = account.getUmsaetze();
            Logger.info(I18N.tr("TransactionSaver.markPreBooked")); //$NON-NLS-1$

            while (umsaetze.hasNext()) {
                final Umsatz umsatz = (Umsatz) umsaetze.next();
                if (umsatz.getFlags() == Umsatz.FLAG_NOTBOOKED) {
                    umsatz.setMeta(TransactionSaver.DELETABLE, TransactionSaver.DELETABLE_TRUE);
                }
            }
        } catch (final RemoteException e) {
            Logger.error(I18N.tr("TransactionSaver.markPreBookedFailed"), e); //$NON-NLS-1$
        }
    }

    /**
     * Sortiert die Liste nach Datum, &auml;lteste zuerst.
     *
     * @param booked
     * @return die sortierte Liste.
     */
    private static List<Umsatz> orderTransactions(final List<Umsatz> booked) {

        // Sort transactions with lambda
        booked.sort((final Umsatz umsatz1, final Umsatz umsatz2) -> {

            int compare = 0;
            try {
                if (umsatz1.getDatum() != null && umsatz2.getDatum() != null) {
                    compare = umsatz1.getDatum().compareTo(umsatz2.getDatum());
                }
            } catch (final RemoteException e) {
                Logger.error(I18N.tr("TransactionSaver.compareTransactionReadDateFailed"), e); //$NON-NLS-1$
            }
            return compare;

        });

        return booked;
    }

    /**
     * L&ouml;scht alle vorgemerkten Ums&auml;tze aus eine Konto.
     *
     * @param account
     *            aus dem alle vorgemerkten Ums&auml;ttze gel&ouml;scht werden
     *            sollen.
     */
    private static void removeNotBooked(final Konto account) {

        try {
            final DBIterator<?> umsaetze = account.getUmsaetze();
            Logger.info(I18N.tr("TransactionSaver.deletePreBooked")); //$NON-NLS-1$

            while (umsaetze.hasNext()) {
                final Umsatz umsatz = (Umsatz) umsaetze.next();
                if (umsatz.getFlags() == Umsatz.FLAG_NOTBOOKED && TransactionSaver.DELETABLE_TRUE
                        .equals(umsatz.getMeta(TransactionSaver.DELETABLE, StringUtils.EMPTY))) {
                    // Nachricht an Hibiscus verschicken
                    // Application.getMessagingFactory().sendMessage(new
                    // ObjectDeletedMessage(umsatz)); // NOPMD
                    umsatz.delete();
                }
            }
        } catch (RemoteException | ApplicationException e) {
            Logger.error(I18N.tr("TransactionSaver.deletePreBookedFailed"), e); //$NON-NLS-1$
        }
    }

    /**
     * Standardkonstruktor
     */
    private TransactionSaver() {

        // Absichtlich leer
    }

}
